@@ -258,9 +258,14 @@ class ChoiceDecoder(AbstractDecoder):
         r = self._createComponent(tagSet, asn1Spec)
         if not decodeFun:
             return r, substrate
-        component, substrate = decodeFun(
-            substrate, r.getComponentTypeMap(), tagSet, length, state
-            )
+        if r.getTagSet() == tagSet: # explicitly tagged Choice
+            component, substrate = decodeFun(
+                substrate, r.getComponentTypeMap()
+                )
+        else:
+            component, substrate = decodeFun(
+                substrate, r.getComponentTypeMap(), tagSet, length, state
+                )
         effectiveTagSet = getattr(
             component, 'getEffectiveTagSet', component.getTagSet
             )()
@@ -340,25 +345,41 @@ class Decoder:
         while state != stStop:
             if state == stDecodeTag:
                 # Decode tag
-                lenOfStream = len(substrate)
-                if lenOfStream < 2:
+                if not substrate:
                     raise error.SubstrateUnderrunError(
-                        'Short octet stream (%d octets)' % lenOfStream
+                        'Short octet stream on tag decoding'
                         )
                 t = ord(substrate[0])
+                tagClass = t&0xC0
+                tagFormat = t&0x20
+                tagId = t&0x1F
+                substrate = substrate[1:]
+                if tagId == 0x1F:
+                    tagId = 0L
+                    while 1:
+                        if not substrate:
+                            raise error.SubstrateUnderrunError(
+                                'Short octet stream on long tag decoding'
+                                )
+                        t = ord(substrate[0])
+                        tagId = tagId << 7 | (t&0x7F)
+                        substrate = substrate[1:]
+                        if not t&0x80:
+                            break
                 lastTag = tag.Tag(
-                    tagClass=(t&0xC0),
-                    tagFormat=(t&0x20),
-                    tagId=t&0x1F
+                    tagClass=tagClass, tagFormat=tagFormat, tagId=tagId
                     )
                 if tagSet is None:
                     tagSet = tag.TagSet((), lastTag) # base tag not recovered
                 else:
                     tagSet = lastTag + tagSet
-                substrate = substrate[1:]
                 state = stDecodeLength
             if state == stDecodeLength:
-                # Decode length (we know there's at least 2 bytes from above)
+                # Decode length
+                if not substrate:
+                     raise error.SubstrateUnderrunError(
+                         'Short octet stream on length decoding'
+                         )
                 firstOctet  = ord(substrate[0])
                 if firstOctet == 128:
                     size = 1
@@ -447,8 +468,9 @@ class Decoder:
                     else:
                         state = stTryAsExplicitTag
             if state == stTryAsExplicitTag:
-                if lastTag[1] == tag.tagFormatConstructed and \
-                       lastTag[0] != tag.tagClassUniversal:
+                if tagSet and \
+                       tagSet[0][1] == tag.tagFormatConstructed and \
+                       tagSet[0][0] != tag.tagClassUniversal:
                     # Assume explicit tagging
                     state = stDecodeTag
                 else:
